#define PYGII_MODULE
#include "pygii.h"

static PyObject*
Event_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
	Event *self;
	
	self = (Event*)type->tp_alloc(type, 0);
	if (self!=NULL) {
		memset(&(self->ev), 0, sizeof(gii_event));
		self->ev.any.size    = 0;
		self->ev.any.type    = evNothing;
		self->ev.any.origin  = GII_EV_ORIGIN_SENDEVENT;
		self->ev.any.error   = 0;
		self->ev.any.target  = GII_EV_TARGET_QUEUE;
		ggCurTime(&self->ev.any.time);
	}
	return (PyObject*)self;
}

static int
Event_init(Event * self, PyObject *args, PyObject *kwds) {
	/* XXX Set atributes wrt kwds */
       	return 0;
}

static PyObject*
Event_getattr(Event *self, char *name) {
	PyObject *a;
	unsigned int i;
	switch(name[0]) {
	case 'e':
		if (!strcmp(name,"error"))
			return PyInt_FromLong((long)self->ev.any.error);
		break;
	case 'o':
		if (!strcmp(name,"origin"))
			return PyInt_FromLong((long)self->ev.any.origin);
		break;
	case 't':
		if (!strcmp(name,"type"))
			return PyInt_FromLong((long)self->ev.any.type);
		if (!strcmp(name,"target"))
			return PyInt_FromLong((long)self->ev.any.target);
		if (!strcmp(name,"time"))
			return PyFloat_FromDouble(TS_TO_FLOAT(&self->ev.any.time));
		break;
	}
	
	switch (self->ev.any.type) {
	case evKeyPress:
	case evKeyRelease:
	case evKeyRepeat:
		if(!strcmp(name,"modifiers"))
			return PyInt_FromLong((long)self->ev.key.modifiers);
		if (!strcmp(name,"symbol"))
			return PyInt_FromLong((long)self->ev.key.sym);
		if (!strcmp(name,"label"))
			return PyInt_FromLong((long)self->ev.key.label);
		if (!strcmp(name,"button"))
			return PyInt_FromLong((long)self->ev.key.button);
		break;
		
	case evPtrRelative:
	case evPtrAbsolute:
		if (!strcmp(name,"x"))
			return PyInt_FromLong((long)self->ev.pmove.x);
		if (!strcmp(name,"y"))
			return PyInt_FromLong((long)self->ev.pmove.y);
		if (!strcmp(name,"z"))
			return PyInt_FromLong((long)self->ev.pmove.z);
		if (!strcmp(name,"wheel"))
			return PyInt_FromLong((long)self->ev.pmove.wheel);
		break;
		
	case evPtrButtonPress:
	case evPtrButtonRelease:
		if (!strcmp(name,"button"))
			return PyInt_FromLong((long)self->ev.pbutton.button);
		break;
		
	case evExpose:
		if (!strcmp(name,"x"))
			return PyInt_FromLong((long)self->ev.expose.x);
		if (!strcmp(name,"y"))
			return PyInt_FromLong((long)self->ev.expose.y);
		if (!strcmp(name,"h"))
			return PyInt_FromLong((long)self->ev.expose.h);
		if (!strcmp(name,"w"))
			return PyInt_FromLong((long)self->ev.expose.w);
		break;
		
	case evCommand:
	case evInformation:
		if (!strcmp(name, "code"))
			return PyInt_FromLong((long)self->ev.cmd.code);
		if (!(self->ev.cmd.code & GII_CMDFLAG_NODATA))
			if (!strcmp(name, "data"))
				return PyString_FromStringAndSize(
					  self->ev.cmd.data,
					  (int)(self->ev.cmd.size -
						sizeof(gii_cmd_nodata_event)));
		break;
		
	case evValRelative:
	case evValAbsolute:
		if (!strcmp(name,"first"))
			return PyInt_FromLong((long)self->ev.val.first);
		if (!strcmp(name,"count"))
			return PyInt_FromLong((long)self->ev.val.count);
		if (!strcmp(name,"value")) {
			a=PyList_New((int)self->ev.val.count);
			if(a) for(i=0;i<self->ev.val.count;i++)
				PyList_SET_ITEM(a,i,
				  PyInt_FromLong((long)self->ev.val.value[i]));
			return a;
		}
		break;
	default:
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return NULL;
}


static int
Event_setattr(Event *self, char *name, PyObject *value) {
	unsigned int i;
	unsigned int n;
	int v[32];
	char * c;
	switch(name[0]) {
	case 'e':
		if (!strcmp(name,"error")) {
			if( pygii_getInt(value, &n) < 0) return -1;
			self->ev.any.error = n;
			return 0;
		}
		break;
	case 'o':
		if (!strcmp(name,"origin"))
			return pygii_getInt(value,&self->ev.any.origin);
		break;
	case 't':
		if (!strcmp(name,"type")) {
			if(pygii_getInt(value,&i)<0) return -1;
			switch(i) {
			case evCommand:
			case evInformation:
			/* size will actually depend on the code itself. */
				self->ev.any.size = sizeof(gii_cmd_event);
				break;
			case evExpose:
				self->ev.any.size = sizeof(gii_expose_event);
				break;
			case evKeyPress:
			case evKeyRelease:
			case evKeyRepeat:
				self->ev.any.size = sizeof(gii_key_event);
				break;
			case evPtrButtonPress:
			case evPtrButtonRelease:
				self->ev.any.size = sizeof(gii_pbutton_event);
				break;
			case evPtrRelative:
			case evPtrAbsolute:
				self->ev.any.size = sizeof(gii_pmove_event);
				break;
			case evValRelative:
			case evValAbsolute:
				self->ev.any.size = sizeof(gii_val_event);
				break;
			default:
				PyErr_SetString(PyExc_ValueError, name);
				return -1;
			}
			self->ev.any.type = i;
			return 0;
		}
		if (!strcmp(name,"target"))
			return pygii_getInt(value, &self->ev.any.target);
		if (!strcmp(name,"time"))
			return pygii_getTime(value, &self->ev.any.time);
		break;
	}
	
	switch (self->ev.any.type) {
	case evKeyPress:
	case evKeyRelease:
	case evKeyRepeat:
		if(!strcmp(name,"modifiers"))
			return pygii_getInt(value,&self->ev.key.modifiers);
		if (!strcmp(name,"symbol"))
			return pygii_getInt(value,&self->ev.key.sym);
		if (!strcmp(name,"label"))
			return pygii_getInt(value,&self->ev.key.label);
		if (!strcmp(name,"button"))
			return pygii_getInt(value,&self->ev.key.button);
		break;
		
	case evPtrRelative:
	case evPtrAbsolute:
		if (!strcmp(name,"x"))
			return pygii_getInt(value,&self->ev.pmove.x);
		if (!strcmp(name,"y"))
			return pygii_getInt(value,&self->ev.pmove.y);
		if (!strcmp(name,"z"))
			return pygii_getInt(value,&self->ev.pmove.z);
		if (!strcmp(name,"wheel"))
			return pygii_getInt(value,&self->ev.pmove.wheel);
		break;
		
	case evPtrButtonPress:
	case evPtrButtonRelease:
		if (!strcmp(name,"button"))
			return pygii_getInt(value,&self->ev.pbutton.button);
		break;
		
	case evExpose:
		if (!strcmp(name,"x"))
			return pygii_getInt(value,&self->ev.expose.x);
		if (!strcmp(name,"y"))
			return pygii_getInt(value,&self->ev.expose.y);
		if (!strcmp(name,"h"))
			return pygii_getInt(value,&self->ev.expose.h);
		if (!strcmp(name,"w"))
			return pygii_getInt(value,&self->ev.expose.w);
		break;
		
	case evCommand:
	case evInformation:
		if (!strcmp(name, "code"))
			return pygii_getInt(value,&self->ev.cmd.code);
		if (!(self->ev.cmd.code & GII_CMDFLAG_NODATA))
			if (!strcmp(name, "data")) {
				n = PyString_Size(value);
				if (PyErr_Occurred())
					return -1;
				if(n>GII_CMD_DATA_MAX) {
					PyErr_SetString(PyExc_ValueError,name);
					return -1;
				}
				c = PyString_AsString(value);
				memmove(self->ev.cmd.data, c, n);
				self->ev.any.size=sizeof(gii_cmd_event)-
					(GII_CMD_DATA_MAX-n);
				return 0;
			}
		break;
		
	case evValRelative:
	case evValAbsolute:
		if (!strcmp(name,"first"))
			return pygii_getInt(value,&self->ev.val.first);

		if (!strcmp(name,"value")) {
			n = PyList_Size(value);
			if (PyErr_Occurred()) return -1;
			if(n>32) {
				PyErr_SetString(PyExc_ValueError, name);
				return -1;
			}
			for(i=0;i<n;i++)
				if(pygii_getInt(value,&v[i])<0)
					return -1;
			for(i=0;i<n;i++)
				self->ev.val.value[i]=v[i];
			self->ev.val.count = n;
			return 0;
		}
		break;
	default:
		break;
	}
	PyErr_SetString(PyExc_AttributeError, name);
	return -1;
}


static char * Event_names[] = {
	"Nothing",
	"Command",
	"Information",
	"Expose",
	"",
	"KeyPress",
	"KeyRelease",
	"KeyRepeat",
	"PtrRelative",
	"PtrAbsolute",
	"PtrButtonPress",
	"PtrButtonRelease",
	"ValRelative",
	"ValAbsolute"
};

static PyObject*
Event_str(Event * self) {
	PyObject * common;
	PyObject * add;
	char _time[50];
	if(!(self->ev.any.type >  0 &&
	     self->ev.any.type <  evLast &&
	     self->ev.any.type != evExpose+1)) {
		return PyString_FromString("Invalid event");
	}
	sprintf(_time,"%f",TS_TO_FLOAT(&(self->ev.any.time)));
	common = PyString_FromFormat(
	    "<size: %2x error: %4x origin: %8x target: %8x time: %s type: %s>",
	    self->ev.any.size,
	    self->ev.any.error,
	    self->ev.any.origin,
	    self->ev.any.target,
	    _time,
	    Event_names[self->ev.any.type]);
	add = NULL;
	switch(self->ev.any.type) {
	case evKeyPress:
	case evKeyRepeat:
	case evKeyRelease:
		add = PyString_FromFormat(
			 " [ modifier:%4x symbol:%4x label:%4x button:%4x ]",
			 self->ev.key.modifiers, self->ev.key.sym,
			 self->ev.key.label, self->ev.key.button);
		break;
	case evValAbsolute:
	case evValRelative:
		add = PyString_FromFormat(" [ first:%8x count:%8x ]",
					  self->ev.val.first,
					  self->ev.val.count);
		break;
	case evPtrAbsolute:
	case evPtrRelative:
		add = PyString_FromFormat(" [ x:%4x y:%4x z:%4x wheel:%4x ]",
					  self->ev.pmove.x,
					  self->ev.pmove.y,
					  self->ev.pmove.z,
					  self->ev.pmove.wheel);
		break;
	case evPtrButtonPress:
	case evPtrButtonRelease:
		add = PyString_FromFormat(" [ button:%4x ]",
					self->ev.pbutton.button);
		break;
	case evExpose:
		add = PyString_FromFormat(" [ x:%4x y:%4x w:%4x h:%4x ]",
					  self->ev.expose.x,self->ev.expose.y,
					  self->ev.expose.w,self->ev.expose.h);
		break;
	case evCommand:
	case evInformation:
		add = PyString_FromFormat(" [ code:%4x ]",
					  self->ev.cmd.code);
		break;
	default:
		break;
	}
	if(add)
		PyString_ConcatAndDel(&common,add);
	return common;
}

static PyObject*
Event_repr(Event * self) {
	PyObject * common;
	PyObject * add;
	char _time[50];

	if(!(self->ev.any.type >  0 &&
	     self->ev.any.type <  evLast &&
	     self->ev.any.type != evExpose+1)) {
		return PyString_FromString("Invalid event");
	}
	sprintf(_time,"%f",TS_TO_FLOAT(&(self->ev.any.time)));
	common = PyString_FromFormat("Event({ 'size':%i, 'error':%i, 'origin':%i, 'target':%i 'time':%s type:%i",
				     self->ev.any.size,
				     self->ev.any.error,
				     self->ev.any.origin,
				     self->ev.any.target,
				     _time,
				     self->ev.any.type);
	
	add = NULL;
	switch(self->ev.any.type) {
	case evKeyPress:
	case evKeyRepeat:
	case evKeyRelease:
		add = PyString_FromFormat(
		   ", 'modifiers':%i, 'symbol':%i, 'label':%i, 'button':%i })",
		   self->ev.key.modifiers, self->ev.key.sym,
		   self->ev.key.label, self->ev.key.button);
		break;
	case evValAbsolute:
	case evValRelative:
		add = PyString_FromFormat(", 'first':%i, 'count':%i })",
					  self->ev.val.first,
					  self->ev.val.count);
		break;
	case evPtrAbsolute:
	case evPtrRelative:
		add = PyString_FromFormat(
			       ", 'x':%i, 'y':%i, 'z':%i, 'wheel':%i })",
			       self->ev.pmove.x,
			       self->ev.pmove.y,
			       self->ev.pmove.z,
			       self->ev.pmove.wheel);
		break;
	case evPtrButtonPress:
	case evPtrButtonRelease:
		add = PyString_FromFormat(", 'button':%i })",
					  self->ev.pbutton.button);
		break;
	case evExpose:
		add = PyString_FromFormat(", 'x':%i, 'y':%i, 'w':%i, 'h':%i })",
					  self->ev.expose.x,self->ev.expose.y,
					  self->ev.expose.w,self->ev.expose.h);
		break;
	case evCommand:
	case evInformation:
		/* XXX Add data if any */
		add=PyString_FromFormat(", 'code':%i })",
					self->ev.cmd.code);
		break;
	default:
		break;
	}
	if (add == NULL)
		add = PyString_FromString(" })");
	PyString_ConcatAndDel(&common,add);
	return common;
}


PyTypeObject EventType = {
	PyObject_HEAD_INIT(NULL)
	0,                          /* ob_size         */
	"GGI._giimodule.Event",     /* tp_name         */
	sizeof(Event),              /* tp_basicsize    */
	0,                          /* tp_itemsize     */
	0,                          /* tp_dealloc      */
	0,                          /* tp_print        */
	(getattrfunc)Event_getattr, /* tp_getattr      */
	(setattrfunc)Event_setattr, /* tp_setattr      */
	0,                          /* tp_compare      */
	(reprfunc)Event_repr,       /* tp_repr         */
	0,                          /* tp_as_number    */
	0,                          /* tp_as_sequence  */
	0,                          /* tp_as_mapping   */
	0,                          /* tp_hash         */
	0,                          /* tp_call         */
	(reprfunc)Event_str,        /* tp_str          */
	0,                          /* tp_getattro     */
	0,                          /* tp_setattro     */
	0,                          /* tp_as_buffer    */
	Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,     /*tp_flags*/
	"Event objects",            /* tp_doc          */
	
	/* 2.0 */
	0,		            /* tp_traverse     */
	0,		            /* tp_clear        */
	/* 2.1 */
	0,		            /* tp_richcompare  */
	0,		            /* tp_weaklistoffset */
	/* 2.2 */
	0,		            /* tp_iter         */
	0,		            /* tp_iternext     */
	0,                          /* tp_methods      */
	0,                          /* tp_members      */
	0,                          /* tp_getset       */
	0,                          /* tp_base         */
	0,                          /* tp_dict         */
	0,                          /* tp_descr_get    */
	0,                          /* tp_descr_set    */
	0,                          /* tp_dictoffset   */
	(initproc)Event_init,       /* tp_init         */
	0,                          /* tp_alloc        */
	Event_new,                  /* tp_new          */
	0,                          /* tp_is_gc        */
	0,                          /* tp_bases        */
        0,                          /* tp_mro          */
	0,                          /* tp_cache        */
	0,                          /* tp_subclasses   */
	0,                          /* tp_weaklist     */
	0                           /* tp_del          */
};
